/*
gcc -std=c99 -Wall -pedantic -Wextra -O3 -march=native -fopenmp \
    -o m-misiurewicz-domains m-misiurewicz-domains.c \
    `PKG_CONFIG_PATH=/path/to/mandelbrot/opt/lib/pkgconfig \
    pkg-config --cflags --libs mandelbrot-graphics`
*/
#include <complex.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <mandelbrot-numerics.h>
#include <mandelbrot-graphics.h>

static inline double cabs2(complex double z) {
  return creal(z) * creal(z) + cimag(z) * cimag(z);
}

extern int main(int argc, char **argv) {
  (void) argc;
  (void) argv;
  int width = 2048;
  int height = 2048;
  int maxiters = 2048;
  double hue = 9.016994374947426e-2;
  complex double cs[5] =
    { -0.75
    , -1.2082320153783807e-01 + I * 8.9693387448293627e-01
    ,  3.2946883886634926e-01 + I * 5.9408266729159365e-01
    , -5.4411945045517796e-01 + I * 6.1529714351441345e-01
    , -1.768666070923659e+00  + I * 3.4365517485742021e-02
    };
  double rs[5] = { 1.5, 0.25, 0.125, 0.125, 4.4194173824159223e-02 };
  m_image *img = m_image_new(width, height);
  for (int w = 0; w < 5; ++w) {
    complex double center = cs[w];
    double radius = rs[w];
    m_d_transform *transform =
      m_d_transform_rectangular(width, height, center, radius);
    for (int per = 1; per <= 5; ++per) {
      #pragma omp parallel for schedule(dynamic, 1)
      for (int j = 0; j < height; ++j) {
        complex double *zs = malloc(maxiters * sizeof(*zs));
        int *ps = malloc(maxiters * sizeof(*ps));
        m_compute_t bias = m_unknown;
        for (int i = 0; i < width; ++i) {
          complex double c = i + I * j;
          complex double dc0 = 1;
          m_d_transform_forward(transform, &c, &dc0);
          double pixelspacing2 = cabs2(dc0);
          double pixelspacing = sqrt(pixelspacing2);
          double de = 0;
          complex double z = c;
          complex double dc = 1;
          complex double zp = c;
          complex double dcp = 1;
          complex double dz = 0;
          int pp = 0;
          double mp2 = 1.0 / 0.0;
          double mz2 = 1.0 / 0.0;
          int np = 0;
          for (int n = 0; n < per; ++n) {
            double z2 = cabs2(z);
            if (z2 < mz2) {
              mz2 = z2;
              int p = n + 1;
              if (bias == m_interior) {
                if (m_d_interior_de(&de, &dz, z, c, p, 64)) {
                  de /= pixelspacing;
                  bias = m_interior;
                  break;
                }
              } else {
                zs[np] = z;
                ps[np] = p;
                np++;
              }
            }
            if (z2 > 65536) {
              de = sqrt(z2 / cabs2(dc * dc0)) * log(z2);
              bias = m_exterior;
              break;
            }
            dc = 2 * z * dc + 1;
            z = z * z + c;
          }
          if (de == 0) {
            for (int n = 0; n < maxiters - per; ++n) {
              double z2 = cabs2(z);
              if (z2 < mz2) {
                mz2 = z2;
                int p = n + 1 + per;
                if (bias == m_interior) {
                  if (m_d_interior_de(&de, &dz, z, c, p, 64)) {
                    de /= pixelspacing;
                    bias = m_interior;
                    break;
                  }
                } else {
                  zs[np] = z;
                  ps[np] = p;
                  np++;
                }
              }
              if (z2 > 65536) {
                de = sqrt(z2 / cabs2(dc * dc0)) * log(z2);
                bias = m_exterior;
                break;
              }
              double p2 = cabs2(z - zp);
              if (p2 < mp2) {
                mp2 = p2;
                pp = n;
              }
              dc = 2 * z * dc + 1;
              z = z * z + c;
              dcp = 2 * zp * dcp + 1;
              zp = zp * zp + c;
            }
          }
          if (de == 0 && bias != m_interior) {
            for (int n = 0; n < np; ++n) {
              if (m_d_interior_de(&de, &dz, zs[n], c, ps[n], 64)) {
                de /= pixelspacing;
                bias = m_interior;
                break;
              }
            }
          }
          double sat = bias == m_interior ? 0 : (pp > 0) * (0.25 + 0.75 /
            pow(1.0 + sqrt(mp2 / pixelspacing2) /
              (1024 * (w == 4 && per != 3 ? 4 : 1)), 2.0));
          m_image_plot(img, i, j,
            m_pixel_hsva((pp - 1) * hue, sat, tanh(de), 1));
        }
        free(zs);
        free(ps);
      }
      m_image_dirty(img);
      char filename[16];
      snprintf(filename, 16, "%c%d.png", 'a' + w, per);
      m_image_save_png(img, filename);
    }
    m_d_transform_delete(transform);
  }
  m_image_delete(img);
  return 0;
}
